import unittest

from reactivex import Observable
from reactivex.disposable import SerialDisposable
from reactivex.operators import filter, filter_indexed
from reactivex.testing import ReactiveTest, TestScheduler, is_prime

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


def test_is_prime():
    assert not is_prime(1)
    assert is_prime(2)
    assert is_prime(3)
    assert not is_prime(4)
    assert is_prime(5)
    assert not is_prime(6)


class TestFilter(unittest.TestCase):
    def test_filter_complete(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
            on_next(610, 12),
            on_error(620, "ex"),
            on_completed(630),
        )

        def create() -> Observable[int]:
            def predicate(x: int) -> bool:
                invoked[0] += 1
                return is_prime(x)

            return xs.pipe(filter(predicate))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(230, 3),
            on_next(340, 5),
            on_next(390, 7),
            on_next(580, 11),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]
        assert invoked[0] == 9

    def test_filter_true(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
        )

        def create() -> Observable[int]:
            def predicate(x: int) -> bool:
                invoked[0] += 1
                return True

            return xs.pipe(filter(predicate))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]
        assert invoked[0] == 9

    def test_filter_false(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
        )

        def create() -> Observable[int]:
            def predicate(x: int) -> bool:
                invoked[0] += 1
                return False

            return xs.pipe(filter(predicate))

        results = scheduler.start(create)

        assert results.messages == [on_completed(600)]
        assert xs.subscriptions == [subscribe(200, 600)]
        assert invoked[0] == 9

    def test_filter_dispose(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
        )

        def create():
            def predicate(x: int) -> bool:
                invoked[0] += 1
                return is_prime(x)

            return xs.pipe(filter(predicate))

        results = scheduler.start(create, disposed=400)
        assert results.messages == [on_next(230, 3), on_next(340, 5), on_next(390, 7)]
        assert xs.subscriptions == [subscribe(200, 400)]
        assert invoked[0] == 5

    def test_filter_error(self):
        scheduler = TestScheduler()
        invoked = [0]
        ex = "ex"
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_error(600, ex),
            on_next(610, 12),
            on_error(620, "ex"),
            on_completed(630),
        )

        def create():
            def predicate(x: int) -> bool:
                invoked[0] += 1
                return is_prime(x)

            return xs.pipe(filter(predicate))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(230, 3),
            on_next(340, 5),
            on_next(390, 7),
            on_next(580, 11),
            on_error(600, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]
        assert invoked[0] == 9

    def test_filter_on_error(self):
        scheduler = TestScheduler()
        invoked = [0]
        ex = "ex"
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
            on_next(610, 12),
            on_error(620, "ex"),
            on_completed(630),
        )

        def create():
            def predicate(x: int) -> bool:
                invoked[0] += 1
                if x > 5:
                    raise Exception(ex)

                return is_prime(x)

            return xs.pipe(filter(predicate))

        results = scheduler.start(create)

        assert results.messages == [on_next(230, 3), on_next(340, 5), on_error(380, ex)]
        assert xs.subscriptions == [subscribe(200, 380)]
        assert invoked[0] == 4

    def test_filter_dispose_in_predicate(self):
        scheduler = TestScheduler()
        invoked = [0]
        ys = [None]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
            on_next(610, 12),
            on_error(620, "ex"),
            on_completed(630),
        )
        results = scheduler.create_observer()
        d = SerialDisposable()

        def action(scheduler, state):
            def predicate(x):
                invoked[0] += 1
                if x == 8:
                    d.dispose()

                return is_prime(x)

            ys[0] = xs.pipe(filter(predicate))
            return ys[0]

        scheduler.schedule_absolute(created, action)

        def action1(scheduler, state):
            d.disposable = ys[0].subscribe(results)

        scheduler.schedule_absolute(subscribed, action1)

        def action2(scheduler, state):
            d.dispose()

        scheduler.schedule_absolute(disposed, action2)

        scheduler.start()
        assert results.messages == [on_next(230, 3), on_next(340, 5), on_next(390, 7)]
        assert xs.subscriptions == [subscribe(200, 450)]
        assert invoked[0] == 6

    def test_filter_indexed_complete(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
            on_next(610, 12),
            on_error(620, "ex"),
            on_completed(630),
        )

        def create():
            def predicate(x, index):
                invoked[0] += 1
                return is_prime(x + index * 10)

            return xs.pipe(filter_indexed(predicate))

        results = scheduler.start(create)
        assert results.messages == [on_next(230, 3), on_next(390, 7), on_completed(600)]
        assert xs.subscriptions == [subscribe(200, 600)]
        assert invoked[0] == 9

    def test_filter_indexed_true(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
        )

        def create():
            def predicate(x, index):
                invoked[0] += 1
                return True

            return xs.pipe(filter_indexed(predicate))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]
        assert invoked[0] == 9

    def test_filter_indexed_false(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
        )

        def create():
            def predicate(x, index):
                invoked[0] += 1
                return False

            return xs.pipe(filter_indexed(predicate))

        results = scheduler.start(create)

        assert results.messages == [on_completed(600)]
        assert xs.subscriptions == [subscribe(200, 600)]
        assert invoked[0] == 9

    def test_filter_indexed_dispose(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
        )

        def create():
            def predicate(x, index):
                invoked[0] += 1
                return is_prime(x + index * 10)

            return xs.pipe(filter_indexed(predicate))

        results = scheduler.start(create, disposed=400)
        assert results.messages == [on_next(230, 3), on_next(390, 7)]
        assert xs.subscriptions == [subscribe(200, 400)]
        assert invoked[0] == 5

    def test_filter_indexed_error(self):
        scheduler = TestScheduler()
        invoked = [0]
        ex = "ex"
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_error(600, ex),
            on_next(610, 12),
            on_error(620, "ex"),
            on_completed(630),
        )

        def create():
            def predicate(x, index):
                invoked[0] += 1
                return is_prime(x + index * 10)

            return xs.pipe(filter_indexed(predicate))

        results = scheduler.start(create)

        assert results.messages == [on_next(230, 3), on_next(390, 7), on_error(600, ex)]
        assert xs.subscriptions == [subscribe(200, 600)]
        assert invoked[0] == 9

    def test_filter_indexed_on_error(self):
        scheduler = TestScheduler()
        invoked = [0]
        ex = "ex"
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
            on_next(610, 12),
            on_error(620, "ex"),
            on_completed(630),
        )

        def create():
            def predicate(x, index):
                invoked[0] += 1
                if x > 5:
                    raise Exception(ex)

                return is_prime(x + index * 10)

            return xs.pipe(filter_indexed(predicate))

        results = scheduler.start(create)
        assert results.messages == [on_next(230, 3), on_error(380, ex)]
        assert xs.subscriptions == [subscribe(200, 380)]
        assert invoked[0] == 4

    def test_filter_indexed_dispose_in_predicate(self):
        scheduler = TestScheduler()
        ys = [None]
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(110, 1),
            on_next(180, 2),
            on_next(230, 3),
            on_next(270, 4),
            on_next(340, 5),
            on_next(380, 6),
            on_next(390, 7),
            on_next(450, 8),
            on_next(470, 9),
            on_next(560, 10),
            on_next(580, 11),
            on_completed(600),
            on_next(610, 12),
            on_error(620, "ex"),
            on_completed(630),
        )
        results = scheduler.create_observer()
        d = SerialDisposable()

        def action1(scheduler, state):
            def predicate(x, index):
                invoked[0] += 1
                if x == 8:
                    d.dispose()

                return is_prime(x + index * 10)

            ys[0] = xs.pipe(filter_indexed(predicate))

        scheduler.schedule_absolute(created, action1)

        def action2(scheduler, state):
            d.disposable = ys[0].subscribe(results)

        scheduler.schedule_absolute(subscribed, action2)

        def action3(scheduler, state):
            d.dispose()

        scheduler.schedule_absolute(disposed, action3)

        scheduler.start()
        assert results.messages == [on_next(230, 3), on_next(390, 7)]
        assert xs.subscriptions == [subscribe(200, 450)]
        assert invoked[0] == 6



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestAsObservable(unittest.TestCase):
    def test_as_observable_hides(self):
        some_observable = reactivex.empty()
        assert some_observable.pipe(ops.as_observable()) != some_observable

    def test_as_observable_never(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.never().pipe(ops.as_observable())

        results = scheduler.start(create)

        assert results.messages == []

    def test_as_observable_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(250))

        def create():
            return xs.pipe(ops.as_observable())

        results = scheduler.start(create).messages
        self.assertEqual(1, len(results))
        assert results[0].value.kind == "C" and results[0].time == 250

    def test_as_observable_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(250, ex))

        def create():
            return xs.pipe(ops.as_observable())

        results = scheduler.start(create).messages
        self.assertEqual(1, len(results))
        assert (
            results[0].value.kind == "E"
            and str(results[0].value.exception) == ex
            and results[0].time == 250
        )

    def test_as_observable_Return(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(220, 2), on_completed(250)
        )

        def create():
            return xs.pipe(ops.as_observable())

        results = scheduler.start(create).messages
        self.assertEqual(2, len(results))
        assert (
            results[0].value.kind == "N"
            and results[0].value.value == 2
            and results[0].time == 220
        )
        assert results[1].value.kind == "C" and results[1].time == 250

    def test_as_observable_isnoteager(self):
        scheduler = TestScheduler()
        subscribed = [False]

        def subscribe(obs, scheduler=None):
            subscribed[0] = True
            disp = scheduler.create_hot_observable(
                on_next(150, 1), on_next(220, 2), on_completed(250)
            ).subscribe(obs)

            def func():
                return disp.dispose()

            return func

        xs = reactivex.create(subscribe)
        xs.pipe(ops.as_observable())
        assert not subscribed[0]

        def create():
            return xs.pipe(ops.as_observable())

        scheduler.start(create)

        assert subscribed[0]



import unittest

import pytest

from reactivex.internal import ArgumentOutOfRangeException
from reactivex.internal.constants import DELTA_ZERO, UTC_ZERO
from reactivex.scheduler import VirtualTimeScheduler


class VirtualSchedulerTestScheduler(VirtualTimeScheduler):
    def add(self, absolute, relative):
        return absolute + relative


class TestVirtualTimeScheduler(unittest.TestCase):
    def test_virtual_now_noarg(self):
        scheduler = VirtualSchedulerTestScheduler()
        assert scheduler.clock == 0.0
        assert scheduler.now == UTC_ZERO

    def test_virtual_now_float(self):
        scheduler = VirtualSchedulerTestScheduler(0.0)
        assert scheduler.clock == 0.0
        assert scheduler.now == UTC_ZERO

    def test_virtual_now_timedelta(self):
        scheduler = VirtualSchedulerTestScheduler(DELTA_ZERO)
        assert scheduler.clock == DELTA_ZERO
        assert scheduler.now == UTC_ZERO

    def test_virtual_now_datetime(self):
        scheduler = VirtualSchedulerTestScheduler(UTC_ZERO)
        assert scheduler.clock == UTC_ZERO
        assert scheduler.now == UTC_ZERO

    def test_virtual_schedule_action(self):
        scheduler = VirtualSchedulerTestScheduler()
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        scheduler.schedule(action)

        scheduler.start()
        assert ran is True

    def test_virtual_schedule_action_error(self):
        scheduler = VirtualSchedulerTestScheduler()

        class MyException(Exception):
            pass

        def action(scheduler, state):
            raise MyException()

        with pytest.raises(MyException):
            scheduler.schedule(action)
            scheduler.start()

    def test_virtual_schedule_sleep_error(self):
        scheduler = VirtualSchedulerTestScheduler()

        with pytest.raises(ArgumentOutOfRangeException):
            scheduler.sleep(-1)

    def test_virtual_schedule_advance_clock_error(self):
        scheduler = VirtualSchedulerTestScheduler()

        with pytest.raises(ArgumentOutOfRangeException):
            scheduler.advance_to(scheduler._clock - 1)



import unittest
from datetime import datetime, timedelta, timezone
from time import sleep

import pytest

from reactivex.scheduler.eventloop import IOLoopScheduler

tornado = pytest.importorskip("tornado")
from tornado import ioloop  # isort: skip


class TestIOLoopScheduler(unittest.TestCase):
    def test_ioloop_schedule_now(self):
        loop = ioloop.IOLoop.instance()
        scheduler = IOLoopScheduler(loop)
        diff = scheduler.now - datetime.fromtimestamp(loop.time(), tz=timezone.utc)
        assert abs(diff) < timedelta(milliseconds=1)

    def test_ioloop_schedule_now_units(self):
        loop = ioloop.IOLoop.instance()
        scheduler = IOLoopScheduler(loop)
        diff = scheduler.now
        sleep(0.1)
        diff = scheduler.now - diff
        assert timedelta(milliseconds=80) < diff < timedelta(milliseconds=180)

    def test_ioloop_schedule_action(self):
        loop = ioloop.IOLoop.instance()

        scheduler = IOLoopScheduler(loop)
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        scheduler.schedule(action)

        def done():
            assert ran is True
            loop.stop()

        loop.call_later(0.1, done)
        loop.start()

    def test_ioloop_schedule_action_due(self):
        loop = ioloop.IOLoop.instance()

        scheduler = IOLoopScheduler(loop)
        starttime = loop.time()
        endtime = None

        def action(scheduler, state):
            nonlocal endtime
            endtime = loop.time()

        scheduler.schedule_relative(0.2, action)

        def done():
            assert endtime is not None
            diff = endtime - starttime
            assert diff > 0.18
            loop.stop()

        loop.call_later(0.3, done)
        loop.start()

    def test_ioloop_schedule_action_cancel(self):
        loop = ioloop.IOLoop.instance()

        ran = False
        scheduler = IOLoopScheduler(loop)

        def action(scheduler, state):
            nonlocal ran
            ran = True

        d = scheduler.schedule_relative(0.01, action)
        d.dispose()

        def done():
            assert ran is False
            loop.stop()

        loop.call_later(0.1, done)
        loop.start()



import unittest

from reactivex.testing.marbles import marbles_testing
from reactivex.testing.reactivetest import ReactiveTest

# from reactivex.scheduler import timeout_scheduler, new_thread_scheduler

# marble sequences to test:
# tested_marbles = '0-1-(10)|', '0|', '(10)-(20)|', '(abc)-|'


# class TestFromToMarbles(unittest.TestCase):

#     def test_new_thread_scheduler(self):
#         stream = Observable.from_marbles(marbles)
#         result = stream.to_blocking().to_marbles()
#         self.assertEqual(result, expected)

#         'this is the default scheduler'
#         self._run_test(tested_marbles, new_thread_scheduler)

# def test_timeout_scheduler(self):
#     self._run_test(tested_marbles, timeout_scheduler)

# def test_timeout_new_thread_scheduler(self):
#     self._run_test(tested_marbles, timeout_scheduler, new_thread_scheduler)

# def test_new_thread_scheduler_timeout(self):
#     self._run_test(tested_marbles, new_thread_scheduler, timeout_scheduler)

# def test_timeout_testscheduler(self):
#     '''the test scheduler uses virtual time => `to_marbles` does not
#        see the original delays.
#     '''
#     expected = [t.replace('-', '') for t in tested_marbles]
#     self._run_test(expected, timeout_scheduler, TestScheduler())

# def test_newthread_testscheduler(self):
#     '''the test scheduler uses virtual time => `to_marbles` does not
#        see the original delays.
#     '''
#     expected = [t.replace('-', '') for t in tested_marbles]
#     self._run_test(expected, new_thread_scheduler, TestScheduler())


class TestTestContext(unittest.TestCase):
    def test_start_with_cold_never(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = cold("----")
            "           012345678901234567890"

            def create():
                return obs

            results = start(create)
            expected = []
            assert results == expected

    def test_start_with_cold_empty(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = cold("------|")
            "           012345678901234567890"

            def create():
                return obs

            results = start(create)
            expected = [ReactiveTest.on_completed(206)]
            assert results == expected

    def test_start_with_cold_normal(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = cold("12--3-|")
            "           012345678901234567890"

            def create():
                return obs

            results = start(create)
            expected = [
                ReactiveTest.on_next(200.0, 12),
                ReactiveTest.on_next(204.0, 3),
                ReactiveTest.on_completed(206.0),
            ]
            assert results == expected

    def test_start_with_cold_no_create_function(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = cold("12--3-|")
            "           012345678901234567890"

            results = start(obs)
            expected = [
                ReactiveTest.on_next(200.0, 12),
                ReactiveTest.on_next(204.0, 3),
                ReactiveTest.on_completed(206.0),
            ]
            assert results == expected

    def test_start_with_hot_never(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = hot("------")
            "          012345678901234567890"

            def create():
                return obs

            results = start(create)
            expected = []
            assert results == expected

    def test_start_with_hot_empty(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = hot("---|")
            "          012345678901234567890"

            def create():
                return obs

            results = start(create)
            expected = [
                ReactiveTest.on_completed(203.0),
            ]
            assert results == expected

    def test_start_with_hot_normal(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = hot("-12--3-|")
            "          012345678901234567890"

            def create():
                return obs

            results = start(create)
            expected = [
                ReactiveTest.on_next(201.0, 12),
                ReactiveTest.on_next(205.0, 3),
                ReactiveTest.on_completed(207.0),
            ]
            assert results == expected

    def test_exp(self):
        with marbles_testing() as (start, cold, hot, exp):
            results = exp("12--3--4--5-|")
            "              012345678901234567890"

            expected = [
                ReactiveTest.on_next(200.0, 12),
                ReactiveTest.on_next(204.0, 3),
                ReactiveTest.on_next(207.0, 4),
                ReactiveTest.on_next(210.0, 5),
                ReactiveTest.on_completed(212.0),
            ]
            assert results == expected

    def test_start_with_hot_and_exp(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = hot("     --3--4--5-|")
            expected = exp("--3--4--5-|")
            "               012345678901234567890"

            def create():
                return obs

            results = start(create)
            assert results == expected

    def test_start_with_cold_and_exp(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = cold("     12--3--4--5-|")
            expected = exp(" 12--3--4--5-|")
            "                012345678901234567890"

            def create():
                return obs

            results = start(create)
            assert results == expected

    def test_start_with_cold_and_exp_group(self):
        with marbles_testing() as (start, cold, hot, exp):
            obs = cold("     12--(3,6.5)----(5,#)")
            expected = exp(" 12--(3,6.5)----(5,#)")
            "                012345678901234567890"

            def create():
                return obs

            results = start(create)
            assert results == expected



import asyncio
from asyncio import Future
from typing import Any, Coroutine, List, TypeVar

import reactivex
from reactivex import Notification, Observable
from reactivex import operators as ops
from reactivex.scheduler.eventloop import AsyncIOScheduler

_T = TypeVar("_T")


def to_async_generator(sentinel: Any = None) -> Coroutine[Any, Any, Future[Any]]:
    loop = asyncio.get_event_loop()
    future = loop.create_future()
    notifications: List[Notification[Any]] = []

    def _to_async_generator(source: Observable[_T]):
        def feeder():
            nonlocal future

            if not notifications or future.done():
                return

            notification = notifications.pop(0)

            if notification.kind == "E":
                future.set_exception(notification.exception)
            elif notification.kind == "C":
                future.set_result(sentinel)
            else:
                future.set_result(notification.value)

        def on_next(value: _T) -> None:
            """Takes on_next values and appends them to the notification queue"""

            notifications.append(value)
            loop.call_soon(feeder)

        source.pipe(ops.materialize()).subscribe(on_next)

        async def gen():
            """Generator producing futures"""
            nonlocal future

            loop.call_soon(feeder)
            future = Future()

            return future

        return gen

    return _to_async_generator


async def go(loop):
    scheduler = AsyncIOScheduler(loop)

    xs = reactivex.from_([x for x in range(10)], scheduler=scheduler)
    gen = xs.pipe(to_async_generator())

    # Wish we could write something like:
    # ys = (x for x in yield from gen())
    while True:
        x = await gen()
        if x is None:
            break
        print(x)


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(go(loop))


if __name__ == "__main__":
    main()



import unittest

from reactivex import Observable, operators
from reactivex.testing import ReactiveTest
from reactivex.testing.subscription import Subscription
from reactivex.testing.testscheduler import TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


class TestConcatMap(unittest.TestCase):
    def test_concat_map_and_flatten_each_item(self):
        scheduler = TestScheduler()
        e1 = scheduler.create_hot_observable(
            on_next(220, 1),
            on_next(300, 3),
            on_next(330, 5),
            on_completed(500),
        )
        e2 = scheduler.create_cold_observable(
            on_next(0, 10), on_next(10, 10), on_next(20, 10), on_completed(30)
        )

        def create_inner(x: int) -> Observable[int]:
            return e2.pipe(operators.map(lambda i: i * x))

        def test_create():
            return e1.pipe(operators.concat_map(create_inner))

        results = scheduler.start(test_create)
        assert results.messages == [
            on_next(220, 10),
            on_next(230, 10),
            on_next(240, 10),
            on_next(300, 30),
            on_next(310, 30),
            on_next(320, 30),
            on_next(330, 50),
            on_next(340, 50),
            on_next(350, 50),
            on_completed(500),
        ]
        assert e1.subscriptions == [Subscription(200, 500)]
        assert e2.subscriptions == [
            Subscription(220, 250),
            Subscription(300, 330),
            Subscription(330, 360),
        ]

    def test_concat_map_many_inner_inner_never_completes(self):
        """should concat_ap many outer to many inner, inner never completes"""
        scheduler = TestScheduler()
        e1 = scheduler.create_hot_observable(
            on_next(210, 1),
            on_next(300, 4),  # take 4 will never complete
            on_next(400, 5),
        )
        e2 = scheduler.create_cold_observable(
            on_next(0, 5), on_next(10, 55), on_next(20, 555)
        )

        def create_inner(x: int) -> Observable[int]:
            return e2.pipe(operators.take(x))

        def test_create():
            return e1.pipe(operators.concat_map(create_inner))

        results = scheduler.start(test_create)
        assert results.messages == [
            on_next(210, 5),
            on_next(300, 5),
            on_next(310, 55),
            on_next(320, 555),
        ]
        assert e1.subscriptions == [Subscription(200, 1000)]
        assert e2.subscriptions == [
            Subscription(210, 210),
            Subscription(300, 1000),  # unsubs when we dispose of e1 at end of test
        ]

    def test_concat_map_finalize_before_next(self):
        """should finalize before moving to the next observable"""
        scheduler = TestScheduler()
        e1 = scheduler.create_hot_observable(
            on_next(210, 2),
            on_next(220, 4),
            on_next(600, 6),
        )
        e2 = scheduler.create_cold_observable(
            on_next(50, 5), on_next(100, 55), on_completed(100)
        )

        def create_inner(x: int) -> Observable[int]:
            return e2.pipe(operators.map(lambda i: i * x))

        def test_create():
            return e1.pipe(operators.concat_map(create_inner))

        results = scheduler.start(test_create)
        assert results.messages == [
            on_next(260, 10),
            on_next(310, 110),
            on_next(310 + 50, 20),
            on_next(410, 220),
            on_next(650, 30),
            on_next(700, 330),
        ]
        assert e1.subscriptions == [Subscription(200, 1000)]
        assert e2.subscriptions == [
            Subscription(210, 310),
            Subscription(310, 410),
            Subscription(600, 700),
        ]

    def test_concat_map_inner_errors(self):
        """should propagate errors if the mapped inner throws"""
        scheduler = TestScheduler()
        e1 = scheduler.create_cold_observable(
            on_next(0, 0),
            on_next(50, 1),
            on_next(100, 2),
        )

        inners = [
            scheduler.create_cold_observable(
                on_next(10, 1), on_next(100, 2), on_completed(100)
            ),
            scheduler.create_cold_observable(
                on_next(10, 50), on_error(80, Exception("no"))
            ),
            scheduler.create_cold_observable(
                on_next(10, 1), on_next(100, 2), on_completed(100)
            ),
        ]

        def create_inner(x: int) -> Observable[int]:
            return inners[x]

        def test_create():
            return e1.pipe(operators.concat_map(create_inner))

        results = scheduler.start(test_create)
        assert results.messages == [
            on_next(210, 1),
            on_next(300, 2),
            on_next(310, 50),
            on_error(380, Exception("no")),
        ]
        assert e1.subscriptions == [Subscription(200, 380)]
        e2, e3, e4 = inners
        assert e2.subscriptions == [
            Subscription(200, 300),
        ]
        assert e3.subscriptions == [
            Subscription(300, 380),
        ]
        assert e4.subscriptions == []

    def test_concat_map_outer_errors(self):
        scheduler = TestScheduler()
        e1 = scheduler.create_hot_observable(
            on_next(210, 2),
            on_next(220, 4),
            on_error(230, Exception("a")),
        )
        e2 = scheduler.create_cold_observable(on_next(0, 5), on_completed(100))

        def create_inner(x: int) -> Observable[int]:
            return e2.pipe(operators.map(lambda i: i * x))

        def test_create():
            return e1.pipe(operators.concat_map(create_inner))

        results = scheduler.start(test_create)

        assert results.messages == [on_next(210, 10), on_error(230, Exception("a"))]
        assert e1.subscriptions == [Subscription(200, 230)]
        assert e2.subscriptions == [
            Subscription(210, 230)
        ]  # should not be any further sub and should unsub from e2 on outer error



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestSample(unittest.TestCase):
    def test_sample_regular(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(230, 3),
            on_next(260, 4),
            on_next(300, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_completed(390),
        )

        def create():
            return xs.pipe(ops.sample(50))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(250, 3),
            on_next(300, 5),
            on_next(350, 6),
            on_next(400, 7),
            on_completed(400),
        ]

    def test_sample_error_in_flight(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(230, 3),
            on_next(260, 4),
            on_next(300, 5),
            on_next(310, 6),
            on_error(330, ex),
        )

        def create():
            return xs.pipe(ops.sample(50))

        results = scheduler.start(create)
        assert results.messages == [on_next(250, 3), on_next(300, 5), on_error(330, ex)]

    def test_sample_empty(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.empty().pipe(ops.sample(0))

        results = scheduler.start(create)
        assert results.messages == [on_completed(200)]

    def test_sample_error(self):
        ex = "ex"
        scheduler = TestScheduler()

        def create():
            return reactivex.throw(ex).pipe(ops.sample(0))

        results = scheduler.start(create)

        assert results.messages == [on_error(200, ex)]

    def test_sample_never(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.never().pipe(ops.sample(1))

        results = scheduler.start(create)
        assert results.messages == []



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestSingleOrDefault(unittest.TestCase):
    def test_single_or_default_async_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(250))

        def create():
            return xs.pipe(ops.single_or_default(None, 0))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 0), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_single_or_default_async_one(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(250)
        )

        def create():
            return xs.pipe(ops.single_or_default(None, 0))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 2), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_single_or_default_async_many(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_next(220, 3), on_completed(250)
        )

        def create():
            return xs.pipe(ops.single_or_default(None, 0))

        res = scheduler.start(create=create)

        def predicate(e):
            return not e is None

        # assert res.messages == [on_error(220, predicate)]
        assert [on_error(220, predicate)] == res.messages
        assert xs.subscriptions == [subscribe(200, 220)]

    def test_single_or_default_async_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(210, ex))

        def create():
            return xs.pipe(ops.single_or_default(None, 0))

        res = scheduler.start(create=create)

        assert res.messages == [on_error(210, ex)]
        assert xs.subscriptions == [subscribe(200, 210)]

    def test_single_or_default_async_predicate(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        )

        def create():
            def predicate(x):
                return x % 2 == 1

            return xs.pipe(ops.single_or_default(predicate, 0))

        res = scheduler.start(create=create)

        def predicate(e):
            return not e is None

        assert [on_error(240, predicate)] == res.messages
        assert xs.subscriptions == [subscribe(200, 240)]

    def test_single_or_default_async_predicate_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(250))

        def create():
            def predicate(x):
                return x % 2 == 1

            return xs.pipe(ops.single_or_default(predicate, 0))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 0), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_single_or_default_async_predicate_one(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        )

        def create():
            def predicate(x):
                return x == 4

            return xs.pipe(ops.single_or_default(predicate, 0))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 4), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_single_or_default_async_predicate_none(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        )

        def create():
            def predicate(x):
                return x > 10

            return xs.pipe(ops.single_or_default(predicate, 0))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(250, 0), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_single_or_default_async_predicate_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(210, ex))

        def create():
            def predicate(x):
                return x > 10

            return xs.pipe(ops.single_or_default(predicate, 0))

        res = scheduler.start(create=create)

        assert res.messages == [on_error(210, ex)]
        assert xs.subscriptions == [subscribe(200, 210)]

    def test_single_or_default_async_predicate_throws(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        )

        def create():
            def predicate(x):
                if x < 4:
                    return False
                else:
                    raise Exception(ex)

            return xs.pipe(ops.single_or_default(predicate, 0))

        res = scheduler.start(create=create)

        assert res.messages == [on_error(230, ex)]
        assert xs.subscriptions == [subscribe(200, 230)]



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


class TestForkJoin(unittest.TestCase):
    def test_fork_join_never_never(self):
        scheduler = TestScheduler()
        e1 = reactivex.never()
        e2 = reactivex.never()

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == []

    def test_fork_join_never_empty(self):
        scheduler = TestScheduler()
        e1 = reactivex.never()
        e2 = reactivex.empty()

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_completed(200)]

    def test_fork_join_never_non_empty(self):
        scheduler = TestScheduler()
        e1 = reactivex.never()
        e2 = scheduler.create_hot_observable(
            [on_next(150, 1), on_next(230, 2), on_completed(300)]
        )

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == []

    def test_fork_join_empty_empty(self):
        scheduler = TestScheduler()
        e1 = reactivex.empty()
        e2 = reactivex.empty()

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_completed(200)]

    def test_fork_join_empty_non_empty(self):
        scheduler = TestScheduler()
        e1 = reactivex.empty()
        e2 = scheduler.create_hot_observable(
            [on_next(150, 1), on_next(230, 2), on_completed(300)]
        )

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_completed(200)]

    def test_fork_join_non_empty_non_empty_right_last(self):
        scheduler = TestScheduler()

        msgs1 = [on_next(150, 1), on_next(215, 2), on_completed(230)]
        msgs2 = [on_next(150, 1), on_next(220, 3), on_completed(240)]

        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_next(240, (2, 3)), on_completed(240)]

    def test_fork_join_non_empty_non_empty_left_last(self):
        scheduler = TestScheduler()

        msgs1 = [on_next(150, 1), on_next(215, 2), on_completed(300)]
        msgs2 = [on_next(150, 1), on_next(220, 3), on_completed(240)]

        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_next(300, (2, 3)), on_completed(300)]

    def test_fork_join_empty_error(self):
        ex = RxException()

        scheduler = TestScheduler()
        e1 = reactivex.empty()
        e2 = scheduler.create_hot_observable(
            [on_next(150, 1), on_next(230, 2), on_error(300, ex)]
        )

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_completed(200)]

    def test_fork_join_never_error(self):
        ex = RxException()

        scheduler = TestScheduler()
        e1 = reactivex.never()
        e2 = scheduler.create_hot_observable(
            [on_next(150, 1), on_next(230, 2), on_error(300, ex)]
        )

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_error(300, ex)]

    def test_fork_join_non_empty_error_left_last(self):
        ex = RxException()

        scheduler = TestScheduler()

        msgs1 = [on_next(150, 1), on_next(250, 2), on_completed(330)]
        msgs2 = [on_next(150, 1), on_next(230, 2), on_error(300, ex)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_error(300, ex)]

    def test_fork_join_non_empty_error_right_last(self):
        ex = RxException()

        scheduler = TestScheduler()

        msgs1 = [on_next(150, 1), on_next(250, 2), on_completed(300)]
        msgs2 = [on_next(150, 1), on_next(230, 2), on_error(330, ex)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_error(330, ex)]

    def test_fork_join_error_error_left_last(self):
        ex = RxException()

        scheduler = TestScheduler()

        msgs1 = [on_next(150, 1), on_next(250, 2), on_error(340, ex)]
        msgs2 = [on_next(150, 1), on_next(230, 2), on_error(330, ex)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_error(330, ex)]

    def test_fork_join_error_error_right_last(self):
        ex = RxException()

        scheduler = TestScheduler()

        msgs1 = [on_next(150, 1), on_next(250, 2), on_error(340, ex)]
        msgs2 = [on_next(150, 1), on_next(230, 2), on_error(370, ex)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)

        results = scheduler.start(lambda: reactivex.fork_join(e1, e2))
        assert results.messages == [on_error(340, ex)]

    def test_fork_join_many(self):
        scheduler = TestScheduler()

        msgs1 = [
            on_next(150, 1),
            on_next(210, 2),
            on_next(230, 3),
            on_next(300, 9),
            on_completed(500),
        ]
        msgs2 = [
            on_next(150, 1),
            on_next(205, 3),
            on_next(220, 7),
            on_next(400, 3),
            on_completed(900),
        ]
        msgs3 = [
            on_next(150, 1),
            on_next(250, 2),
            on_next(300, 3),
            on_next(400, 9),
            on_next(500, 2),
            on_completed(850),
        ]
        msgs4 = [
            on_next(150, 1),
            on_next(400, 2),
            on_next(550, 10),
            on_next(560, 11),
            on_next(600, 3),
            on_completed(605),
        ]
        msgs5 = [
            on_next(150, 1),
            on_next(201, 3),
            on_next(550, 10),
            on_next(560, 11),
            on_next(600, 3),
            on_next(900, 99),
            on_completed(905),
        ]

        xs = [
            scheduler.create_hot_observable(x)
            for x in [msgs1, msgs2, msgs3, msgs4, msgs5]
        ]
        results = scheduler.start(lambda: reactivex.fork_join(*xs))
        assert results.messages == [on_next(905, (9, 3, 2, 3, 99)), on_completed(905)]

    def test_fork_join_many_ops(self):
        scheduler = TestScheduler()

        msgs1 = [
            on_next(150, 1),
            on_next(210, 2),
            on_next(230, 3),
            on_next(300, 9),
            on_completed(500),
        ]
        msgs2 = [
            on_next(150, 1),
            on_next(205, 3),
            on_next(220, 7),
            on_next(400, 3),
            on_completed(900),
        ]
        msgs3 = [
            on_next(150, 1),
            on_next(250, 2),
            on_next(300, 3),
            on_next(400, 9),
            on_next(500, 2),
            on_completed(850),
        ]
        msgs4 = [
            on_next(150, 1),
            on_next(400, 2),
            on_next(550, 10),
            on_next(560, 11),
            on_next(600, 3),
            on_completed(605),
        ]
        msgs5 = [
            on_next(150, 1),
            on_next(201, 3),
            on_next(550, 10),
            on_next(560, 11),
            on_next(600, 3),
            on_next(900, 99),
            on_completed(905),
        ]

        xs = [scheduler.create_hot_observable(x) for x in [msgs2, msgs3, msgs4, msgs5]]

        def create():
            return scheduler.create_hot_observable(msgs1).pipe(ops.fork_join(*xs))

        results = scheduler.start(create)
        assert results.messages == [on_next(905, (9, 3, 2, 3, 99)), on_completed(905)]



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler


class TestToArray(ReactiveTest, unittest.TestCase):
    def test_toiterable_completed(self):
        scheduler = TestScheduler()
        msgs = [
            self.on_next(110, 1),
            self.on_next(220, 2),
            self.on_next(330, 3),
            self.on_next(440, 4),
            self.on_next(550, 5),
            self.on_completed(660),
        ]
        xs = scheduler.create_hot_observable(msgs)

        def create():
            return xs.pipe(
                ops.to_iterable(),
                ops.map(list),
            )

        results = scheduler.start(create=create).messages

        assert len(results) == 2
        assert results[0].time == 660
        assert results[0].value.kind == "N"
        assert results[0].value.value == [2, 3, 4, 5]
        assert self.on_completed(660).equals(results[1])
        assert xs.subscriptions == [self.subscribe(200, 660)]

    def test_toiterableerror(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs = [
            self.on_next(110, 1),
            self.on_next(220, 2),
            self.on_next(330, 3),
            self.on_next(440, 4),
            self.on_next(550, 5),
            self.on_error(660, ex),
        ]
        xs = scheduler.create_hot_observable(msgs)

        def create():
            return xs.pipe(ops.to_iterable())

        results = scheduler.start(create=create).messages
        assert results == [self.on_error(660, ex)]
        assert xs.subscriptions == [self.subscribe(200, 660)]

    def test_toiterabledisposed(self):
        scheduler = TestScheduler()
        msgs = [
            self.on_next(110, 1),
            self.on_next(220, 2),
            self.on_next(330, 3),
            self.on_next(440, 4),
            self.on_next(550, 5),
        ]
        xs = scheduler.create_hot_observable(msgs)

        def create():
            return xs.pipe(ops.to_iterable())

        results = scheduler.start(create=create).messages
        assert results == []
        assert xs.subscriptions == [self.subscribe(200, 1000)]



import os
import threading
import unittest
from datetime import timedelta
from time import sleep

import pytest

from reactivex.internal.basic import default_now
from reactivex.scheduler import TimeoutScheduler

CI = os.getenv("CI") is not None


class TestTimeoutScheduler(unittest.TestCase):
    def test_timeout_singleton(self):
        scheduler = [TimeoutScheduler(), TimeoutScheduler.singleton()]
        assert scheduler[0] is scheduler[1]

        gate = [threading.Semaphore(0), threading.Semaphore(0)]
        scheduler = [None, None]

        def run(idx):
            scheduler[idx] = TimeoutScheduler()
            gate[idx].release()

        for idx in (0, 1):
            threading.Thread(target=run, args=(idx,)).start()
            gate[idx].acquire()

        assert scheduler[0] is not None
        assert scheduler[1] is not None
        assert scheduler[0] is scheduler[1]

    def test_timeout_extend(self):
        class MyScheduler(TimeoutScheduler):
            pass

        scheduler = [
            MyScheduler(),
            MyScheduler.singleton(),
            TimeoutScheduler.singleton(),
        ]
        assert scheduler[0] is scheduler[1]
        assert scheduler[0] is not scheduler[2]

    @pytest.mark.skipif(CI, reason="Flaky test in GitHub Actions")
    def test_timeout_now(self):
        scheduler = TimeoutScheduler()
        diff = scheduler.now - default_now()
        assert abs(diff) < timedelta(milliseconds=1)

    @pytest.mark.skipif(CI, reason="Flaky test in GitHub Actions")
    def test_timeout_now_units(self):
        scheduler = TimeoutScheduler()
        diff = scheduler.now
        sleep(1.1)
        diff = scheduler.now - diff
        assert timedelta(milliseconds=1000) < diff < timedelta(milliseconds=1300)

    def test_timeout_schedule_action(self):
        scheduler = TimeoutScheduler()
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        scheduler.schedule(action)

        sleep(0.1)
        assert ran is True

    def test_timeout_schedule_action_due(self):
        scheduler = TimeoutScheduler()
        starttime = default_now()
        endtime = None

        def action(scheduler, state):
            nonlocal endtime
            endtime = default_now()

        scheduler.schedule_relative(timedelta(milliseconds=200), action)

        sleep(0.4)
        assert endtime is not None
        diff = endtime - starttime
        assert diff > timedelta(milliseconds=180)

    def test_timeout_schedule_action_cancel(self):
        ran = False
        scheduler = TimeoutScheduler()

        def action(scheduler, state):
            nonlocal ran
            ran = True

        d = scheduler.schedule_relative(timedelta(milliseconds=300), action)
        d.dispose()

        sleep(0.1)
        assert ran is False



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestWindowWithCount(unittest.TestCase):
    def test_window_with_count_basic(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(280, 4),
            on_next(320, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_completed(600),
        )

        def create():
            def proj(w, i):
                return w.pipe(ops.map(lambda x: str(i) + " " + str(x)))

            return xs.pipe(
                ops.window_with_count(3, 2), ops.map_indexed(proj), ops.merge_all()
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, "0 2"),
            on_next(240, "0 3"),
            on_next(280, "0 4"),
            on_next(280, "1 4"),
            on_next(320, "1 5"),
            on_next(350, "1 6"),
            on_next(350, "2 6"),
            on_next(380, "2 7"),
            on_next(420, "2 8"),
            on_next(420, "3 8"),
            on_next(470, "3 9"),
            on_completed(600),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]

    def test_window_with_count_disposed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(280, 4),
            on_next(320, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_completed(600),
        )

        def create():
            def proj(w, i):
                return w.pipe(ops.map(lambda x: str(i) + " " + str(x)))

            return xs.pipe(
                ops.window_with_count(3, 2), ops.map_indexed(proj), ops.merge_all()
            )

        results = scheduler.start(create, disposed=370)
        assert results.messages == [
            on_next(210, "0 2"),
            on_next(240, "0 3"),
            on_next(280, "0 4"),
            on_next(280, "1 4"),
            on_next(320, "1 5"),
            on_next(350, "1 6"),
            on_next(350, "2 6"),
        ]
        assert xs.subscriptions == [subscribe(200, 370)]

    def test_window_with_count_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(280, 4),
            on_next(320, 5),
            on_next(350, 6),
            on_next(380, 7),
            on_next(420, 8),
            on_next(470, 9),
            on_error(600, ex),
        )

        def create():
            def mapper(w, i):
                def mapping(x):
                    return "%s %s" % (i, x)

                return w.pipe(ops.map(mapping))

            return xs.pipe(
                ops.window_with_count(3, 2), ops.map_indexed(mapper), ops.merge_all()
            )

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, "0 2"),
            on_next(240, "0 3"),
            on_next(280, "0 4"),
            on_next(280, "1 4"),
            on_next(320, "1 5"),
            on_next(350, "1 6"),
            on_next(350, "2 6"),
            on_next(380, "2 7"),
            on_next(420, "2 8"),
            on_next(420, "3 8"),
            on_next(470, "3 9"),
            on_error(600, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 600)]



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestFind(unittest.TestCase):
    def test_find_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1))

        def create():
            return xs.pipe(ops.find(lambda x, i, s: True))

        res = scheduler.start(create)

        assert res.messages == []

    def test_find_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(210))

        def create():
            return xs.pipe(ops.find(lambda x, i, s: True))

        res = scheduler.start(create)

        assert res.messages == [on_next(210, None), on_completed(210)]

    def test_find_single(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(220)
        )

        def create():
            return xs.pipe(ops.find(lambda x, i, s: x == 2))

        res = scheduler.start(create)

        assert res.messages == [on_next(210, 2), on_completed(210)]

    def test_find_notfound(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(220)
        )

        def create():
            return xs.pipe(ops.find(lambda x, i, s: x == 3))

        res = scheduler.start(create)

        assert res.messages == [on_next(220, None), on_completed(220)]

    def test_find_Error(self):
        ex = Exception("error")
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_error(220, ex)
        )

        def create():
            return xs.pipe(ops.find(lambda x, i, s: x == 3))

        res = scheduler.start(create)

        assert res.messages == [on_error(220, ex)]

    def test_find_throws(self):
        ex = "error"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(220)
        )

        def create():
            def predicate(x, i, source):
                raise Exception(ex)

            return xs.pipe(ops.find(predicate))

        res = scheduler.start(create)

        assert res.messages == [on_error(210, ex)]



import unittest

import reactivex
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestGenerateWithRelativeTime(unittest.TestCase):
    def test_generate_timespan_finite(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: x <= 3, lambda x: x + 1, lambda x: x + 1
            )

        results = scheduler.start(create)
        assert results.messages == [
            on_next(201, 0),
            on_next(203, 1),
            on_next(206, 2),
            on_next(210, 3),
            on_completed(210),
        ]

    def test_generate_timespan_throw_condition(self):
        ex = "ex"
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: _raise(ex), lambda x: x + 1, lambda x: x + 1
            )

        results = scheduler.start(create)
        assert results.messages == [on_error(200, ex)]

    def test_generate_timespan_throw_iterate(self):
        ex = "ex"
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: True, lambda x: _raise(ex), lambda x: x + 1
            )

        results = scheduler.start(create)
        assert results.messages == [on_next(201, 0), on_error(201, ex)]

    def test_generate_timespan_throw_timemapper(self):
        ex = "ex"
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: True, lambda x: x + 1, lambda x: _raise(ex)
            )

        results = scheduler.start(create)
        assert results.messages == [on_error(200, ex)]

    def test_generate_timespan_dispose(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: True, lambda x: x + 1, lambda x: x + 1
            )

        results = scheduler.start(create, disposed=210)
        assert results.messages == [on_next(201, 0), on_next(203, 1), on_next(206, 2)]

    def test_generate_datetime_offset_finite(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: x <= 3, lambda x: x + 1, lambda x: x + 1
            )

        results = scheduler.start(create)
        assert results.messages == [
            on_next(201, 0),
            on_next(203, 1),
            on_next(206, 2),
            on_next(210, 3),
            on_completed(210),
        ]

    def test_generate_datetime_offset_throw_condition(self):
        ex = "ex"
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: _raise(ex), lambda x: x + 1, lambda x: x + 1
            )

        results = scheduler.start(create)
        assert results.messages == [on_error(200, ex)]

    def test_generate_datetime_offset_throw_iterate(self):
        ex = "ex"
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: True, lambda x: _raise(ex), lambda x: x + 1
            )

        results = scheduler.start(create)
        assert results.messages == [on_next(201, 0), on_error(201, ex)]

    def test_generate_datetime_offset_throw_time_mapper(self):
        ex = "ex"
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: True, lambda x: x + 1, lambda x: _raise(ex)
            )

        results = scheduler.start(create)
        assert results.messages == [on_error(200, ex)]

    def test_generate_datetime_offset_dispose(self):
        scheduler = TestScheduler()

        def create():
            return reactivex.generate_with_relative_time(
                0, lambda x: True, lambda x: x + 1, lambda x: x + 1
            )

        results = scheduler.start(create, disposed=210)
        assert results.messages == [on_next(201, 0), on_next(203, 1), on_next(206, 2)]



import asyncio
import os
import threading
import unittest
from datetime import datetime, timedelta, timezone

import pytest

from reactivex.scheduler.eventloop import AsyncIOThreadSafeScheduler

CI = os.getenv("CI") is not None


class TestAsyncIOThreadSafeScheduler(unittest.TestCase):
    @pytest.mark.skipif(CI, reason="Flaky test in GitHub Actions")
    def test_asyncio_threadsafe_schedule_now(self):
        loop = asyncio.get_event_loop()
        scheduler = AsyncIOThreadSafeScheduler(loop)
        diff = scheduler.now - datetime.fromtimestamp(loop.time(), tz=timezone.utc)
        assert abs(diff) < timedelta(milliseconds=2)

    @pytest.mark.skipif(CI, reason="Flaky test in GitHub Actions")
    def test_asyncio_threadsafe_schedule_now_units(self):
        loop = asyncio.get_event_loop()
        scheduler = AsyncIOThreadSafeScheduler(loop)
        diff = scheduler.now
        yield from asyncio.sleep(0.1)
        diff = scheduler.now - diff
        assert timedelta(milliseconds=80) < diff < timedelta(milliseconds=180)

    def test_asyncio_threadsafe_schedule_action(self):
        loop = asyncio.get_event_loop()

        async def go():
            scheduler = AsyncIOThreadSafeScheduler(loop)
            ran = False

            def action(scheduler, state):
                nonlocal ran
                ran = True

            def schedule():
                scheduler.schedule(action)

            threading.Thread(target=schedule).start()

            await asyncio.sleep(0.1)
            assert ran is True

        loop.run_until_complete(go())

    def test_asyncio_threadsafe_schedule_action_due(self):
        loop = asyncio.get_event_loop()

        async def go():
            scheduler = AsyncIOThreadSafeScheduler(loop)
            starttime = loop.time()
            endtime = None

            def action(scheduler, state):
                nonlocal endtime
                endtime = loop.time()

            def schedule():
                scheduler.schedule_relative(0.2, action)

            threading.Thread(target=schedule).start()

            await asyncio.sleep(0.3)
            assert endtime is not None
            diff = endtime - starttime
            assert diff > 0.18

        loop.run_until_complete(go())

    def test_asyncio_threadsafe_schedule_action_cancel(self):
        loop = asyncio.get_event_loop()

        async def go():
            ran = False
            scheduler = AsyncIOThreadSafeScheduler(loop)

            def action(scheduler, state):
                nonlocal ran
                ran = True

            def schedule():
                d = scheduler.schedule_relative(0.05, action)
                d.dispose()

            threading.Thread(target=schedule).start()

            await asyncio.sleep(0.3)
            assert ran is False

        loop.run_until_complete(go())

    def cancel_same_thread_common(self, test_body):
        update_state = {"ran": False, "dispose_completed": False}

        def action(scheduler, state):
            update_state["ran"] = True

        # Make the actual test body run in deamon thread, so that in case of
        # failure it doesn't hang indefinitely.
        def thread_target():
            loop = asyncio.new_event_loop()
            scheduler = AsyncIOThreadSafeScheduler(loop)

            test_body(scheduler, action, update_state)

            async def go():
                await asyncio.sleep(0.2)

            loop.run_until_complete(go())

        thread = threading.Thread(target=thread_target)
        thread.daemon = True
        thread.start()
        thread.join(0.3)
        assert update_state["dispose_completed"] is True
        assert update_state["ran"] is False

    def test_asyncio_threadsafe_cancel_non_relative_same_thread(self):
        def test_body(scheduler, action, update_state):
            d = scheduler.schedule(action)

            # Test case when dispose is called on thread on which loop is not
            # yet running, and non-relative schedele is used.
            d.dispose()
            update_state["dispose_completed"] = True

        self.cancel_same_thread_common(test_body)

    def test_asyncio_threadsafe_schedule_action_cancel_same_thread(self):
        def test_body(scheduler, action, update_state):
            d = scheduler.schedule_relative(0.05, action)

            # Test case when dispose is called on thread on which loop is not
            # yet running, and relative schedule is used.
            d.dispose()
            update_state["dispose_completed"] = True

        self.cancel_same_thread_common(test_body)

    def test_asyncio_threadsafe_schedule_action_cancel_same_loop(self):
        def test_body(scheduler, action, update_state):
            d = scheduler.schedule_relative(0.1, action)

            def do_dispose():
                d.dispose()
                update_state["dispose_completed"] = True

            # Test case when dispose is called in loop's callback.
            scheduler._loop.call_soon(do_dispose)

        self.cancel_same_thread_common(test_body)



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestExpand(unittest.TestCase):
    def test_expand_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_completed(300))

        def create():
            def mapper():
                return scheduler.create_cold_observable(
                    on_next(100, 1), on_next(200, 2), on_completed(300)
                )

            return xs.pipe(ops.expand(mapper))

        results = scheduler.start(create)

        assert results.messages == [on_completed(300)]
        assert xs.subscriptions == [subscribe(200, 300)]

    def test_expand_error(self):
        scheduler = TestScheduler()
        ex = "ex"
        xs = scheduler.create_hot_observable(on_error(300, ex))

        def create():
            def mapper(x):
                return scheduler.create_cold_observable(
                    on_next(100 + x, 2 * x),
                    on_next(200 + x, 3 * x),
                    on_completed(300 + x),
                )

            return xs.pipe(ops.expand(mapper))

        results = scheduler.start(create)

        assert results.messages == [on_error(300, ex)]
        assert xs.subscriptions == [subscribe(200, 300)]

    def test_expand_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable()

        def create():
            def mapper(x):
                return scheduler.create_cold_observable(
                    on_next(100 + x, 2 * x),
                    on_next(200 + x, 3 * x),
                    on_completed(300 + x),
                )

            return xs.pipe(ops.expand(mapper))

        results = scheduler.start(create)

        assert results.messages == []
        assert xs.subscriptions == [subscribe(200, 1000)]

    def test_expand_basic(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(550, 1), on_next(850, 2), on_completed(950)
        )

        def create():
            def mapper(x):
                return scheduler.create_cold_observable(
                    on_next(100, 2 * x), on_next(200, 3 * x), on_completed(300)
                )

            return xs.pipe(ops.expand(mapper))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(550, 1),
            on_next(650, 2),
            on_next(750, 3),
            on_next(750, 4),
            on_next(850, 2),
            on_next(850, 6),
            on_next(850, 6),
            on_next(850, 8),
            on_next(950, 9),
            on_next(950, 12),
            on_next(950, 4),
            on_next(950, 12),
            on_next(950, 12),
            on_next(950, 16),
        ]
        assert xs.subscriptions == [subscribe(200, 950)]

    def test_expand_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(550, 1), on_next(850, 2), on_completed(950)
        )

        def create():
            def mapper(x):
                raise Exception(ex)

            return xs.pipe(ops.expand(mapper))

        results = scheduler.start(create)

        assert results.messages == [on_next(550, 1), on_error(550, ex)]
        assert xs.subscriptions == [subscribe(200, 550)]



import threading
from datetime import timedelta
from time import sleep

import pytest

from reactivex.internal.basic import default_now
from reactivex.scheduler.mainloop import QtScheduler

QtCore = pytest.importorskip("PyQt5.QtCore")


@pytest.fixture(scope="module")
def app():
    # share qt application among all tests
    app = QtCore.QCoreApplication([])
    yield app
    # teardown


class TestQtSchedulerPyQt5:
    def test_pyqt5_schedule_now(self):
        scheduler = QtScheduler(QtCore)
        diff = scheduler.now - default_now()
        assert abs(diff) < timedelta(milliseconds=1)

    def test_pyqt5_schedule_now_units(self):
        scheduler = QtScheduler(QtCore)
        diff = scheduler.now
        sleep(0.1)
        diff = scheduler.now - diff
        assert timedelta(milliseconds=80) < diff < timedelta(milliseconds=180)

    def test_pyqt5_schedule_action(self, app):

        scheduler = QtScheduler(QtCore)
        gate = threading.Semaphore(0)
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        scheduler.schedule(action)

        def done():
            app.quit()
            gate.release()

        QtCore.QTimer.singleShot(50, done)
        app.exec_()

        gate.acquire()
        assert ran is True

    def test_pyqt5_schedule_action_due_relative(self, app):

        scheduler = QtScheduler(QtCore)
        gate = threading.Semaphore(0)
        starttime = default_now()
        endtime = None

        def action(scheduler, state):
            nonlocal endtime
            endtime = default_now()

        scheduler.schedule_relative(0.2, action)

        def done():
            app.quit()
            gate.release()

        QtCore.QTimer.singleShot(300, done)
        app.exec_()

        gate.acquire()
        assert endtime is not None
        diff = endtime - starttime
        assert diff > timedelta(milliseconds=180)

    def test_pyqt5_schedule_action_due_absolute(self, app):

        scheduler = QtScheduler(QtCore)
        gate = threading.Semaphore(0)
        starttime = default_now()
        endtime = None

        def action(scheduler, state):
            nonlocal endtime
            endtime = default_now()

        scheduler.schedule_absolute(starttime + timedelta(seconds=0.2), action)

        def done():
            app.quit()
            gate.release()

        QtCore.QTimer.singleShot(300, done)
        app.exec_()

        gate.acquire()
        assert endtime is not None
        diff = endtime - starttime
        assert diff > timedelta(milliseconds=180)

    def test_pyqt5_schedule_action_cancel(self, app):

        ran = False
        scheduler = QtScheduler(QtCore)
        gate = threading.Semaphore(0)

        def action(scheduler, state):
            nonlocal ran
            ran = True

        d = scheduler.schedule_relative(0.1, action)
        d.dispose()

        def done():
            app.quit()
            gate.release()

        QtCore.QTimer.singleShot(300, done)
        app.exec_()

        gate.acquire()
        assert ran is False

    def test_pyqt5_schedule_action_periodic(self, app):

        scheduler = QtScheduler(QtCore)
        gate = threading.Semaphore(0)
        period = 0.050
        counter = 3

        def action(state):
            nonlocal counter
            if state:
                counter -= 1
                return state - 1

        scheduler.schedule_periodic(period, action, counter)

        def done():
            app.quit()
            gate.release()

        QtCore.QTimer.singleShot(300, done)
        app.exec_()

        gate.acquire()
        assert counter == 0

    def test_pyqt5_schedule_periodic_cancel(self, app):

        scheduler = QtScheduler(QtCore)
        gate = threading.Semaphore(0)
        period = 0.05
        counter = 3

        def action(state):
            nonlocal counter
            if state:
                counter -= 1
                return state - 1

        disp = scheduler.schedule_periodic(period, action, counter)

        def dispose():
            disp.dispose()

        QtCore.QTimer.singleShot(100, dispose)

        def done():
            app.quit()
            gate.release()

        QtCore.QTimer.singleShot(300, done)
        app.exec_()

        gate.acquire()
        assert 0 < counter < 3



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestFinally(unittest.TestCase):
    def test_finally_only_called_once_empty(self):
        invasserte_count = [0]

        def action():
            invasserte_count[0] += 1
            return invasserte_count

        some_observable = reactivex.empty().pipe(ops.finally_action(action))

        d = some_observable.subscribe()
        d.dispose()
        d.dispose()
        self.assertEqual(1, invasserte_count[0])

    def test_finally_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(250))
        invasserted = [False]

        def create():
            def action():
                invasserted[0] = True
                return invasserted[0]

            return xs.pipe(ops.finally_action(action))

        results = scheduler.start(create).messages
        self.assertEqual(1, len(results))
        assert results[0].value.kind == "C" and results[0].time == 250
        assert invasserted[0]

    def test_finally_return(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(250)
        )
        invasserted = [False]

        def create():
            def action():
                invasserted[0] = True
                return invasserted[0]

            return xs.pipe(ops.finally_action(action))

        results = scheduler.start(create).messages
        self.assertEqual(2, len(results))
        assert (
            results[0].value.kind == "N"
            and results[0].time == 210
            and results[0].value.value == 2
        )
        assert results[1].value.kind == "C" and results[1].time == 250
        assert invasserted[0]

    def test_finally_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(250, ex))
        invasserted = [False]

        def create():
            def action():
                invasserted[0] = True
                return invasserted[0]

            return xs.pipe(ops.finally_action(action))

        results = scheduler.start(create).messages
        self.assertEqual(1, len(results))
        assert (
            results[0].value.kind == "E"
            and results[0].time == 250
            and str(results[0].value.exception) == ex
        )
        assert invasserted[0]



import unittest
from typing import Optional

from reactivex import Observable
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


class TestDistinctUntilChanged(unittest.TestCase):
    def test_default_if_empty_non_empty1(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(280, 42),
            on_next(360, 43),
            on_completed(420),
        )

        def create() -> Observable[Optional[int]]:
            return xs.pipe(ops.default_if_empty())

        results = scheduler.start(create)

        assert results.messages == [
            on_next(280, 42),
            on_next(360, 43),
            on_completed(420),
        ]
        assert xs.subscriptions == [subscribe(200, 420)]

    def test_default_if_empty_non_empty2(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(280, 42), on_next(360, 43), on_completed(420)
        )

        def create():
            return xs.pipe(ops.default_if_empty(-1))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(280, 42),
            on_next(360, 43),
            on_completed(420),
        ]
        assert xs.subscriptions == [subscribe(200, 420)]

    def test_default_if_empty_empty1(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_completed(420))

        def create():
            return xs.pipe(ops.default_if_empty(None))

        results = scheduler.start(create)

        assert results.messages == [on_next(420, None), on_completed(420)]
        assert xs.subscriptions == [subscribe(200, 420)]

    def test_default_if_empty_empty2(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_completed(420))

        def create():
            return xs.pipe(ops.default_if_empty(-1))

        results = scheduler.start(create)

        assert results.messages == [on_next(420, -1), on_completed(420)]
        assert xs.subscriptions == [subscribe(200, 420)]



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestSum(unittest.TestCase):
    def test_sum_int32_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(250))

        def create():
            return xs.pipe(ops.sum())

        res = scheduler.start(create=create).messages
        assert res == [on_next(250, 0), on_completed(250)]

    def test_sum_int32_return(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(250)
        )

        def create():
            return xs.pipe(ops.sum())

        res = scheduler.start(create=create).messages
        assert res == [on_next(250, 2), on_completed(250)]

    def test_sum_int32_some(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_completed(250),
        )

        def create():
            return xs.pipe(ops.sum())

        res = scheduler.start(create=create).messages
        assert res == [on_next(250, 2 + 3 + 4), on_completed(250)]

    def test_sum_int32_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(210, ex))

        def create():
            return xs.pipe(ops.sum())

        res = scheduler.start(create=create).messages
        assert res == [on_error(210, ex)]

    def test_sum_int32_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1))

        def create():
            return xs.pipe(ops.sum())

        res = scheduler.start(create=create).messages
        assert res == []

    def test_sum_mapper_regular_int32(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(210, "fo"),
            on_next(220, "b"),
            on_next(230, "qux"),
            on_completed(240),
        )

        def create():
            return xs.pipe(ops.sum(lambda x: len(x)))

        res = scheduler.start(create=create)

        assert res.messages == [on_next(240, 6), on_completed(240)]
        assert xs.subscriptions == [subscribe(200, 240)]



import unittest
from datetime import datetime, timezone

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas
def _raise(ex):
    raise RxException(ex)


class TestTimeout(unittest.TestCase):
    def test_timeout_in_time(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(230, 3),
            on_next(260, 4),
            on_next(300, 5),
            on_next(350, 6),
            on_completed(400),
        )

        def create():
            return xs.pipe(ops.timeout(500, None))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(210, 2),
            on_next(230, 3),
            on_next(260, 4),
            on_next(300, 5),
            on_next(350, 6),
            on_completed(400),
        ]

    def test_timeout_out_of_time(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1),
            on_next(210, 2),
            on_next(230, 3),
            on_next(260, 4),
            on_next(300, 5),
            on_next(350, 6),
            on_completed(400),
        )

        def create():
            return xs.pipe(ops.timeout(205))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(210, 2),
            on_next(230, 3),
            on_next(260, 4),
            on_next(300, 5),
            on_next(350, 6),
            on_completed(400),
        ]

    def test_timeout_timeout_occurs_1(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 1),
            on_next(130, 2),
            on_next(310, 3),
            on_next(400, 4),
            on_completed(500),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, -1), on_next(200, -2), on_next(310, -3), on_completed(320)
        )

        def create():
            return xs.pipe(ops.timeout(100, ys))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(350, -1),
            on_next(500, -2),
            on_next(610, -3),
            on_completed(620),
        ]
        assert xs.subscriptions == [subscribe(200, 300)]
        assert ys.subscriptions == [subscribe(300, 620)]

    def test_timeout_timeout_occurs_2(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 1),
            on_next(130, 2),
            on_next(240, 3),
            on_next(310, 4),
            on_next(430, 5),
            on_completed(500),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, -1), on_next(200, -2), on_next(310, -3), on_completed(320)
        )

        def create():
            return xs.pipe(ops.timeout(100, ys))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(240, 3),
            on_next(310, 4),
            on_next(460, -1),
            on_next(610, -2),
            on_next(720, -3),
            on_completed(730),
        ]
        assert xs.subscriptions == [subscribe(200, 410)]
        assert ys.subscriptions == [subscribe(410, 730)]

    def test_timeout_timeout_occurs_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 1),
            on_next(130, 2),
            on_next(240, 3),
            on_next(310, 4),
            on_next(430, 5),
            on_completed(500),
        )
        ys = scheduler.create_cold_observable()

        def create():
            return xs.pipe(ops.timeout(100, ys))

        results = scheduler.start(create)

        assert results.messages == [on_next(240, 3), on_next(310, 4)]
        assert xs.subscriptions == [subscribe(200, 410)]
        assert ys.subscriptions == [subscribe(410, 1000)]

    def test_timeout_timeout_occurs_completed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_completed(500))
        ys = scheduler.create_cold_observable(on_next(100, -1))

        def create():
            return xs.pipe(ops.timeout(100, ys))

        results = scheduler.start(create)

        assert results.messages == [on_next(400, -1)]
        assert xs.subscriptions == [subscribe(200, 300)]
        assert ys.subscriptions == [subscribe(300, 1000)]

    def test_timeout_timeout_occurs_error(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_error(500, "ex"))
        ys = scheduler.create_cold_observable(on_next(100, -1))

        def create():
            return xs.pipe(ops.timeout(100, ys))

        results = scheduler.start(create)

        assert results.messages == [on_next(400, -1)]
        assert xs.subscriptions == [subscribe(200, 300)]
        assert ys.subscriptions == [subscribe(300, 1000)]

    def test_timeout_timeout_not_occurs_completed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_completed(250))
        ys = scheduler.create_cold_observable(on_next(100, -1))

        def create():
            return xs.pipe(ops.timeout(100, ys))

        results = scheduler.start(create)

        assert results.messages == [on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]
        assert ys.subscriptions == []

    def test_timeout_timeout_not_occurs_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_error(250, ex))
        ys = scheduler.create_cold_observable(on_next(100, -1))

        def create():
            return xs.pipe(ops.timeout(100, ys))

        results = scheduler.start(create)

        assert results.messages == [on_error(250, ex)]
        assert xs.subscriptions == [subscribe(200, 250)]
        assert ys.subscriptions == []

    def test_timeout_timeout_does_not_occur(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 1),
            on_next(130, 2),
            on_next(240, 3),
            on_next(320, 4),
            on_next(410, 5),
            on_completed(500),
        )
        ys = scheduler.create_cold_observable(
            on_next(50, -1), on_next(200, -2), on_next(310, -3), on_completed(320)
        )

        def create():
            return xs.pipe(ops.timeout(100, ys))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(240, 3),
            on_next(320, 4),
            on_next(410, 5),
            on_completed(500),
        ]
        assert xs.subscriptions == [subscribe(200, 500)]
        assert ys.subscriptions == []

    def test_timeout_datetime_offset_timeout_occurs(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(410, 1))
        ys = scheduler.create_cold_observable(on_next(100, -1))

        def create():
            return xs.pipe(
                ops.timeout(datetime.fromtimestamp(400, tz=timezone.utc), ys)
            )

        results = scheduler.start(create)

        assert results.messages == [on_next(500, -1)]
        assert xs.subscriptions == [subscribe(200, 400)]
        assert ys.subscriptions == [subscribe(400, 1000)]

    def test_timeout_datetime_offset_timeout_does_not_occur_completed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(310, 1), on_completed(390))
        ys = scheduler.create_cold_observable(on_next(100, -1))

        def create():
            return xs.pipe(
                ops.timeout(datetime.fromtimestamp(400, tz=timezone.utc), ys)
            )

        results = scheduler.start(create)

        assert results.messages == [on_next(310, 1), on_completed(390)]
        assert xs.subscriptions == [subscribe(200, 390)]
        assert ys.subscriptions == []

    def test_timeout_datetime_offset_timeout_does_not_occur_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(310, 1), on_error(390, ex))
        ys = scheduler.create_cold_observable(on_next(100, -1))

        def create():
            return xs.pipe(
                ops.timeout(datetime.fromtimestamp(400, tz=timezone.utc), ys)
            )

        results = scheduler.start(create)

        assert results.messages == [on_next(310, 1), on_error(390, ex)]
        assert xs.subscriptions == [subscribe(200, 390)]
        assert ys.subscriptions == []

    def test_timeout_datetime_offset_timeout_occur_2(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(310, 1), on_next(350, 2), on_next(420, 3), on_completed(450)
        )
        ys = scheduler.create_cold_observable(on_next(100, -1))

        def create():
            return xs.pipe(
                ops.timeout(datetime.fromtimestamp(400, tz=timezone.utc), ys)
            )

        results = scheduler.start(create)

        assert results.messages == [on_next(310, 1), on_next(350, 2), on_next(500, -1)]
        assert xs.subscriptions == [subscribe(200, 400)]
        assert ys.subscriptions == [subscribe(400, 1000)]

    def test_timeout_datetime_offset_timeout_occur_3(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(310, 1), on_next(350, 2), on_next(420, 3), on_completed(450)
        )
        ys = scheduler.create_cold_observable()

        def create():
            return xs.pipe(
                ops.timeout(datetime.fromtimestamp(400, tz=timezone.utc), ys)
            )

        results = scheduler.start(create)

        assert results.messages == [on_next(310, 1), on_next(350, 2)]
        assert xs.subscriptions == [subscribe(200, 400)]
        assert ys.subscriptions == [subscribe(400, 1000)]



import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


class TestTakeUntil(unittest.TestCase):
    def test_take_until_preempt_somedata_next(self):
        scheduler = TestScheduler()
        left_msgs = [
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        ]
        right_msgs = [on_next(150, 1), on_next(225, 99), on_completed(230)]
        left = scheduler.create_hot_observable(left_msgs)
        right = scheduler.create_hot_observable(right_msgs)

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2), on_next(220, 3), on_completed(225)]

    def test_take_until_preempt_somedata_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        left_msgs = [
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        ]
        right_msgs = [on_next(150, 1), on_error(225, ex)]
        left = scheduler.create_hot_observable(left_msgs)
        right = scheduler.create_hot_observable(right_msgs)

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2), on_next(220, 3), on_error(225, ex)]

    def test_take_until_nopreempt_somedata_empty(self):
        scheduler = TestScheduler()
        left_msgs = [
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        ]
        right_msgs = [on_next(150, 1), on_completed(225)]
        left = scheduler.create_hot_observable(left_msgs)
        right = scheduler.create_hot_observable(right_msgs)

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        ]

    def test_take_until_nopreempt_somedata_never(self):
        scheduler = TestScheduler()
        left_msgs = [
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        ]
        left = scheduler.create_hot_observable(left_msgs)
        right = reactivex.never()

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)
        assert results.messages == [
            on_next(210, 2),
            on_next(220, 3),
            on_next(230, 4),
            on_next(240, 5),
            on_completed(250),
        ]

    def test_take_until_preempt_never_next(self):
        scheduler = TestScheduler()
        right_msgs = [on_next(150, 1), on_next(225, 2), on_completed(250)]
        left = reactivex.never()
        right = scheduler.create_hot_observable(right_msgs)

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)
        assert results.messages == [on_completed(225)]

    def test_take_until_preempt_never_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        right_msgs = [on_next(150, 1), on_error(225, ex)]
        left = reactivex.never()
        right = scheduler.create_hot_observable(right_msgs)

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)
        assert results.messages == [on_error(225, ex)]

    def test_take_until_nopreempt_never_empty(self):
        scheduler = TestScheduler()
        right_msgs = [on_next(150, 1), on_completed(225)]
        left = reactivex.never()
        right = scheduler.create_hot_observable(right_msgs)

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)
        assert results.messages == []

    def test_take_until_nopreempt_never_never(self):
        scheduler = TestScheduler()
        left = reactivex.never()
        right = reactivex.never()

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)
        assert results.messages == []

    def test_take_until_preempt_beforefirstproduced(self):
        scheduler = TestScheduler()
        left_msgs = [on_next(150, 1), on_next(230, 2), on_completed(240)]
        right_msgs = [on_next(150, 1), on_next(210, 2), on_completed(220)]
        l = scheduler.create_hot_observable(left_msgs)
        r = scheduler.create_hot_observable(right_msgs)

        def create():
            return l.pipe(ops.take_until(r))

        results = scheduler.start(create)
        assert results.messages == [on_completed(210)]

    def test_take_until_preempt_beforefirstproduced_remain_silent_and_proper_disposed(
        self,
    ):
        scheduler = TestScheduler()
        left_msgs = [on_next(150, 1), on_error(215, "ex"), on_completed(240)]
        right_msgs = [on_next(150, 1), on_next(210, 2), on_completed(220)]
        source_not_disposed = [False]

        def action():
            source_not_disposed[0] = True

        left = scheduler.create_hot_observable(left_msgs).pipe(
            ops.do_action(on_next=action)
        )

        right = scheduler.create_hot_observable(right_msgs)

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)

        assert results.messages == [on_completed(210)]
        assert not source_not_disposed[0]

    def test_take_until_nopreempt_afterlastproduced_proper_disposed_signal(self):
        scheduler = TestScheduler()
        left_msgs = [on_next(150, 1), on_next(230, 2), on_completed(240)]
        right_msgs = [on_next(150, 1), on_next(250, 2), on_completed(260)]
        signal_not_disposed = [False]
        left = scheduler.create_hot_observable(left_msgs)

        def action():
            signal_not_disposed[0] = True

        right = scheduler.create_hot_observable(right_msgs).pipe(
            ops.do_action(on_next=action)
        )

        def create():
            return left.pipe(ops.take_until(right))

        results = scheduler.start(create)
        assert results.messages == [on_next(230, 2), on_completed(240)]
        assert not signal_not_disposed[0]



# Helpers.
# Run this cell always after kernel restarts. All other cells are autonomous.
from __future__ import print_function

import inspect
import logging

# getting the current thread
import threading
import time
from random import randint

import reactivex

logging.basicConfig(format="%(threadName)s:%(message)s")
log = logging.getLogger("Rx")
log.setLevel(logging.WARNING)

sleep, now = time.sleep, time.time

ts_glob = 0  # global start time


def reset_start_time(show_doc_for=None, title=True, sleep=None):
    "resets global start time and also prints doc strings"
    global ts_glob
    if sleep:
        log("main thread sleeping %ss" % sleep)
        time.sleep(sleep)
    ts_glob, d = time.time(), show_doc_for
    if title:
        if d:
            if title == True:
                title = d.__name__
        header(title)
    if not d:
        return
    # print the function sig and doc if given
    # py2 compatible way:
    deco, fdef = inspect.getsource(d).split("def ", 1)
    fdef = "def ".join((deco, fdef.split(")", 1)[0])) + "):"
    print(
        "module %s\n%s\n%s\n%s"
        % (
            d.__module__,
            fdef.strip(),
            ("    " + (d.__doc__ or "n.a.").strip()),
            "-" * 80,
        )
    )


rst = reset_start_time


def log(*msg):
    s = " ".join([str(s) for s in msg])
    print("%s %s %s" % (dt(ts_glob), cur_thread(), s))


def header(msg):
    print("\n\n%s %s %s\n" % ("=" * 10, msg, "=" * 10))


def rand():
    return randint(100, 999)


def to_int(s):
    return int(s) if s.isdigit() else s


def dt(ts):
    # the time delta of now to given ts (in millis, 1 float)
    return str("%.1f" % ((time.time() - ts) * 1000)).rjust(6)


class ItemGetter:
    "allows to throw an object onto a format string"

    def __init__(self, obj):
        self.obj = obj

    def __getitem__(self, k, d=None):
        return getattr(self.obj, k, d)


class Subscriber:
    def __init__(self, observed_stream, **kw):
        print("")
        name = kw.get("name", str(hash(self))[-5:])
        log(
            "New subscription (%s) on stream" % str(name).strip(), hash(observed_stream)
        )
        self.ts = time.time()  # tstart, for dts at events
        # no postifx after name, sometimes it ends with '\n':
        self.name = name

    def _on(self, what, v=""):
        print(
            "%s %s [%s] %s: %s -> %s"
            % (dt(ts_glob), cur_thread(), what, dt(self.ts), v, self.name)
        )

    def on_next(self, v):
        return self._on("next", v)

    def on_error(self, v):
        return self._on("err ", v)

    def on_completed(self):
        return self._on("cmpl", "fin")


def subs(src, **kw):
    # required for e.g. .multicast:
    obs = Subscriber(src, **kw)
    subscription = src.subscribe(obs)
    if kw.pop("return_subscriber", None):
        return subscription, obs
    return subscription


threads = []


def cur_thread():
    def _cur():
        "return a unique number for the current thread"
        n = threading.current_thread().name
        if "Main" in n:
            return "    M"
        return "%5s" % ("T" + n.rsplit("-", 1)[-1])

    # you could show all running threads via this:
    # threads = ' '.join([t.name for t in threading.enumerate()])
    # return '%s of %s' % (_cur(), threads)
    return _cur()


def marble_stream(s):
    return reactivex.from_marbles(s).to_blocking()
